package o1.robots.tribal

import scala.util.parsing.combinator.JavaTokenParsers
import scala.util.Random

// NOTE TO STUDENTS: You are not required to understand the code in this file.
// It is used by the Tribe class to turn Strings that contain RoboSpeak into
// Line and Instruction objects.

private[tribal] class RoboSpeakGrammar(val program: Tribe) extends JavaTokenParsers:
  import program.Instruction.*

  def parse(input: String): Line =
    this.parseAll(this.line, input) match
      case Success(parsedCode, _) => parsedCode
      case Failure(problem, _)    => throw TribeFileException(s"Invalid RoboSpeak. Message from the parser:\n$problem")
      case Error(problem, _)      => throw TribeFileException(s"Invalid RoboSpeak. Message from the parser:\n$problem")

  def line = labelDef | instruction | directive | empty

  // Instructions:
  def instruction = (
    move        | spin        | uturn       | waitInstr   | switch      |
    talk        | ifempty     | ifnempty    | iffriend    | ifnfriend   |
    ifwall      | ifnwall     | ifrandom    | goto        | callsub     |
    returnInstr | ifeq        | ifneq       | ifgt        | iflt        |
    set         | enemiesnear | friendsnear | foddernear  | fodderleft  |
    shout       | score       | add1        | plus        | minus       |
    enemiesdir  | friendsdir  )                    ^^ ( Line(_, None, None) )
  def        move = "move"                         ^^ ( _ => Move() )
  def        spin = "spin"                         ^^ ( _ => Spin() )
  def       uturn = "uturn"                        ^^ ( _ => UTurn() )
  def   waitInstr = "wait"                         ^^ ( _ => Wait() )
  def      switch = "switch"                       ^^ ( _ => Switch() )
  def        talk = "talk"~>param                  ^^ ( line => Talk(line) )
  def     ifempty = "ifempty"~>param               ^^ ( line => IfEmpty(line) )
  def    ifnempty = "ifnempty"~>param              ^^ ( line => IfNEmpty(line) )
  def    iffriend = "iffriend"~>param              ^^ ( line => IfFriend(line) )
  def   ifnfriend = "ifnfriend"~>param             ^^ ( line => IfNFriend(line) )
  def      ifwall = "ifwall"~>param                ^^ ( line => IfWall(line) )
  def     ifnwall = "ifnwall"~>param               ^^ ( line => IfNWall(line) )
  def    ifrandom = "ifrandom"~>param              ^^ ( line => IfRandom(line) )
  def        goto = "goto"~>param                  ^^ ( line => Goto(line) )
  def        ifeq = "ifeq"~>param~param~param      ^^ { case n1~n2~line => IfEq(n1, n2, line) }
  def       ifneq = "ifneq"~>param~param~param     ^^ { case n1~n2~line => IfNEq(n1, n2, line) }
  def        ifgt = "ifgt"~>param~param~param      ^^ { case n1~n2~line => IfGT(n1, n2, line) }
  def        iflt = "iflt"~>param~param~param      ^^ { case n1~n2~line => IfLT(n1, n2, line) }
  def     callsub = "callsub"~>param               ^^ ( line => CallSub(line) )
  def returnInstr = "return"                       ^^ ( _ => Return() )
  def         set = "set"~>slotParam~param         ^^ { case slot~value => SetSlot(slot.left.getOrElse(""), value) }
  def enemiesnear = "enemiesnear"                  ^^ ( _ => EnemiesNear() )
  def friendsnear = "friendsnear"                  ^^ ( _ => FriendsNear() )
  def  foddernear = "foddernear"                   ^^ ( _ => FodderNear() )
  def  fodderleft = "fodderleft"                   ^^ ( _ => FodderLeft() )
  def       score = "score"                        ^^ ( _ => Score() )
  def  enemiesdir = "enemiesdir"~>param            ^^ ( relDir => EnemiesDir(relDir) )
  def  friendsdir = "friendsdir"~>param            ^^ ( relDir => FriendsDir(relDir) )
  def       shout = "shout"~>param                 ^^ ( line => Shout(line) )
  def        add1 = "add1"~>slotParam              ^^ ( slot => Add1(slot.left.getOrElse("")) )
  def        plus = "plus"~>slotParam~param~param  ^^ { case slot1~value1~value2 => Plus(slot1.left.getOrElse(""), value1, value2) }
  def       minus = "minus"~>slotParam~param~param ^^ { case slot1~value1~value2 => Minus(slot1.left.getOrElse(""), value1, value2) }

  // Parameters of instructions:
  def      param = intParam | slotParam | labelParam
  def  slotParam = Tribe.Slots.mkString("|").r ^^ ( Left(_) )
  def labelParam = raw"\S+".r                  ^^ ( Left(_) )
  def   intParam = wholeNumber                 ^^ ( numString => Right(numString.toInt) )

  // Labels, empty lines, and directives
  def       empty = ""                         ^^ ( _ => Line(NoCommand(), None, None) )
  def    labelDef = raw"\S+:".r                ^^ ( labelName => Line(LabelDefinition(), Some(labelName.init), None) )
  def   directive = "pacifist"                 ^^ ( directiveName => Line(Directive(), None, Some(directiveName)) )

end RoboSpeakGrammar

private[tribal] case class Line(val instruction: Instruction, val label: Option[String], val directive: Option[String])

